
import numpy as np
from scipy.stats import gaussian_kde
import tables
import pandas as pd

import yaml
import os

from matplotlib import pyplot as plt
from textwrap import dedent

from .. import BaseModel
from ..doctools import DocClass


def result_filename(setup, path='.'):
    return os.path.join(path, str(setup) + '.result.yml')


class BaseResult(DocClass):
    """
    A base class for documentary result classes. See usage in example/model2.py
    """
    @classmethod
    def describe(cls, setup=None, **kwargs) -> str:
        with open(result_filename(setup)) as f:
            data = yaml.safe_load(f)
            return super().describe(setup, **data)

    threshold = None
    def calculate_threshold(self, n_min=30):
        """
        Calculates a rejection criteria for given data.
        Use this is only for explorative studies, for a real GLUE approach,
        you may not change an a priori defined rejection criterium.

        :param n_min: Minimum number of behavioural runs
        """
        # Calculate NSE threshold
        threshold = 0.7
        n = 0
        while np.sum(np.array(self.data.cols.like1[:]) > threshold) < n_min and threshold > -2:
            threshold -= 0.05
        n = np.sum(np.array(self.data.cols.like1[:]) > threshold)
        return threshold, n

    def __init__(self, model: BaseModel, result_file: str = None, outputdir = '.'):
        self.name = model.name
        if not hasattr(self, 'result_filename'):
            self.result_filename = result_file or f'{self.name}.h5'

        self.data_file = tables.open_file(self.result_filename)
        self.data = self.data_file.get_node(f'/{model}')
        self.model = model
        # Calculate the behavioural model
        self.threshold, self.n = self.calculate_threshold()
        self.obs = self.model.data.Q.to_pandas()
        self.outputdir = outputdir

    def save(self, verbose):
        self.dotty_plot()
        self.timeseries_plot()
        data = self.result_summary()
        if verbose:
            print(data)
        with open(self.filename('result', 'yml'), 'w') as f:
            yaml.safe_dump(data, f)

    def prune_results(self, condition='like1>=0.0'):
        """
        Prunes the result table
        :param condition: A Python condition to select behaving runs
        :return:
        """
        new_filename = self.result_filename.replace('.h5','.pruned.h5')
        pruned_data = self.data.read_where(condition)
        with tables.open_file(new_filename, 'w') as o:
            tab = tables.Table(o.root, self.name, self.data.description)
            tab.append(pruned_data)
        self.result_filename = new_filename
        self.data_file = tables.open_file(self.result_filename)
        self.data = self.data_file.get_node(f'/{self.model}')

    def close(self):
        self.data_file.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, value, traceback):
        self.close()

    def filename(self, *names):
        ext = '.'.join(names)
        return f'{self.outputdir}/{self.name}.{ext}'

    def load_obs(self):
        df = pd.read_csv('daten/glauburg_temp.csv', index_col=[0], parse_dates=True)
        return df.Q

    def dotty_plot(self):
        """
        Plots the distribution of parameters in the behavioural runs
        """
        take = np.array(self.data.cols.like1[:] > self.threshold)
        fig = plt.figure(figsize=(8, 8), dpi=100)
        p_names = [cn[3:] for cn in self.data.colnames if cn[:3] == 'par']
        letters = [chr(ord('a') + i) for i in range(20)]
        for i, pn in enumerate(p_names):
            plt.subplot(((len(p_names)-1) // 3) + 1, 3, i + 1)
            params = self.data.col(f'par{pn}')[take]
            plt.plot(params, self.data.col('like1')[take], 'x')
            plt.title(pn.replace('_', ' '), loc='left', fontsize=10)
            try:
                gk = gaussian_kde(params)
                x = np.linspace(params.min(), params.max(), 1001)
                plt.twinx()
                plt.plot(x, gk(x), 'r:')
                plt.yticks([])
            except (np.linalg.linalg.LinAlgError, ValueError):
                pass
        plt.subplots_adjust(left=0.075, bottom=0.05, right=0.95, top=0.95, wspace=0.2, hspace=0.4)
        fig.savefig(self.filename('dotty', 'png'))
        return self.filename('dotty', 'png')

    def timeseries_plot(self):
        """
        Plots the result timeseries with the p5 to p95 uncertainty interval
        """
        import pylab as plt
        nse = np.array(self.data.cols.like1)
        best_run = nse.argmax()
        best = self.data.cols.simulation[best_run]

        take = nse > self.threshold
        data = self.data.cols.simulation[:][take].data

        p5 = np.percentile(data, 5, 0)
        p95 = np.percentile(data, 95, 0)

        fig = plt.figure(figsize=(16, 8), dpi=100)
        time = np.arange(np.datetime64('1991', 'D'), np.datetime64('2019', 'D'))
        plt.fill_between(time, p5, p95, facecolor='#ffff00', edgecolor='none', label='Modelled uncertainty')
        plt.plot(time, self.obs, 'k-', label='Observed')
        plt.plot(time, best, 'r-', label='Best model')
        ax = plt.gca()
        ax.xaxis_date()
        ax.set_xlim(np.datetime64(f'{self.model.validation_start}-01-01'))
        plt.axvline(np.datetime64(f'{self.model.validation_start}-01-01'), ls='--', c='k', lw=3, alpha=0.5)
        plt.title('{} NSE>{:0.2f}, n={}'.format(self.name.capitalize(), self.threshold, take.sum()),
                  fontsize=24)
        plt.legend()

        fig.savefig(self.filename('timeseries', 'png'))
        return self.filename('timeseries', 'png')

    def like(self, row):
        """Returns the 4 objective functions for a model run"""
        return [self.data.col(colname)[row] for colname in self.data.colnames if colname.startswith('like')]

    def best_run_id(self):
        """Returns the row number of the best run"""
        return np.array(self.data.cols.like1[:]).argmax()

    def result_summary(self) -> dict:
        best_run_id = self.best_run_id()
        like = self.like(best_run_id)
        return dict(
            n=int(self.n),
            threshold=float(self.threshold),
            best_run_id=int(best_run_id),
            NSE_c=float(like[0]),
            NSE_v=float(like[1]),
            PBIAS_c=float(like[2]),
            PBIAS_v=float(like[3]),
        )


from .basemodel import get_model_class as _get_model_class
from .basemodel import BaseModel as _BaseModel


def gui(model):
    """
    Opens a gui to run and explore the model
    Usage: cmflumped doc <model.py>
           <model.py>: The Python file containing the model
    """
    from cmflumped.gui import GUI
    m = _get_model_class(model)()
    with GUI(m) as mgui:
        mgui.show()


def result(*setups):
    """
    Loads and analyses the result file from cmf.lumped run model.py 10000
    """
    from .doctools.result import BaseResult
    for model in setups:
        m = _get_model_class(model)()
        r: BaseResult
        Result = _get_model_class(model, 'Result')
        with Result(m, '.') as r:
            r.save(verbose=True)


def doc(*setups, in_browser=False):
    """
    Creates documentation files for the model
    Usage: cmflumped doc <model.py>
           <model.py>: The Python file containing the model
    """
    import matplotlib as mpl
    mpl.use('Agg')
    from cmflumped.doctools import Documentation
    # Schreibe die rst-Dateien
    for setup in setups:
        model = _get_model_class(setup)()
        docu = Documentation(model)
        docu.make_rst()
        # Erzeuge die HTML-Dokumentation
        build_dir = docu.compile_html()

    if in_browser:
        import webbrowser
        webbrowser.open((build_dir / 'html' / 'index.html').as_uri())


def descr(model):
    """
    Prints out a description of your model
    Usage: cmflumped descr <model.py>
           <model.py>: The Python file containing the model
    """
    import cmf
    m = _get_model_class(model)()
    print(cmf.describe(m.project))


def run(model, runs=None, sampler='lhs'):
    """
    Runs the model
    Usage: cmf.lumped run <model.py> [runs] [sampler]
           <model.py>: The Python file containing the model
           [runs]: Number of runs (default=1)
           [sampler]: spotpy sampler to use,
                eg. mc (Monte Carlo), lhs (latin hypercube sampling), dds - see spotpy documentation

    Example:
        Single run:
            cmf.lumped run example/model1.py
        Many runs:
            cmf.lumped run example/model1.py 1000 lhs
    """
    from cmflumped.spotpy_helper import sample
    m: _BaseModel = _get_model_class(model)()
    m.verbose = not runs
    if runs:
        n = int(runs)
        sample(m, n, sampler, save_threshold=0.0)
    else:
        from spotpy.parameter import create_set
        p = create_set(m, 'optguess')
        m.simulation(p)




import networkx as nx
import cmf
import numpy as np
import matplotlib.axes
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation


class Network:

    def __init__(self, outlet: cmf.flux_node):
        self.nodes = {}
        self.graph = nx.Graph()
        self.add_node(outlet)
        self.__replace_abstract_nodes_with_concrete_storages(outlet.project)

    def __replace_abstract_nodes_with_concrete_storages(self, project: cmf.project):
        """
        self.nodes contains only abstract cmf.flux_nodes.
        This method replaces the nodes with its concrete storages
        """
        storages = {s.node_id: s for s in project.get_storages()}
        self.nodes.update(storages)

    def __node_pos(self, node: cmf.flux_node):
        """
        Gets the position of the node in a 2D view
        Uses displacement keywords to make the view nicer
        :param node: A CMF flux node
        :return: A [x, z] array
        """
        displacement_dict = dict(
            Snow=(+1, 0),
            Canopy=(-1, +1),
            Evaporation=(-1, -18),
            Transpiration=(-2, -18),
            Rainfall=(0, -17.5)
        )
        pos = np.array(node.position)[[0, -1]]
        for s in displacement_dict:
            if str(node).startswith('{' + s):
                pos += np.array(displacement_dict[s])
        return pos

    def add_node(self, node: cmf.flux_node):
        """
        Adds a cmf node and all of its connected nodes into a networkx graph
        :param node: A cmf flux node
        :param G: A networkx.Graph object. If None, a new Graph is created
        :param nodes_done: A set of handled nodes, if None, a new one is created
        :return: The graph
        """
        self.graph.add_node(
            node.node_id,
            position=self.__node_pos(node),
            name=node.Name.replace(' of cell #0', '')
        )

        self.nodes[node.node_id] = node

        for c in node.connections:
            target = c.get_target(node)
            self.graph.add_edge(
                c.left_node().node_id, c.right_node().node_id,
                label=c.short_string().split('#')[0])
            if target.node_id not in self.nodes:
                self.add_node(target)

    def get_volume(self, node_id):
        """
        Returns the stored volume of a node
        :param node_id: A flux node id.
        :return: the stored water volume in m³
        """
        node = self.nodes[node_id]
        if hasattr(node, 'volume'):
            return max(node.volume, 0.0)
        else:
            return 0.0

    def get_volumes(self):
        """
        Returns a list of volumes for each node of the graph
        :return:
        """
        return [self.get_volume(n) for n in self.graph.nodes]

    def get_connection(self, left_id, right_id)->cmf.flux_connection:
        return self.nodes[left_id].connection_to(self.nodes[right_id])

    def get_flux(self, t, left_id, right_id):
        c = self.get_connection(left_id, right_id)
        return np.abs(c.q(c.left_node(), t))

    def get_fluxes(self, t):
        """
        Returns a list a fluxes for each edge in the graph
        :param t: current time to query the fluxes
        :return:
        """
        return [
            self.get_flux(t, *e)
            for e in self.graph.edges
        ]


class Fluxogram:

    def __init__(self, outlet: cmf.flux_node, axis=None, *, with_edge_labels=False):
        self.network = Network(outlet)
        if axis is None:
            axis = plt.gca()
        self.axis: matplotlib.axes.Axes() = axis
        self.axis.set_axis_off()
        self.with_edge_labels = with_edge_labels
        self.node_labels = {}
        self.node_markers = []
        self.edge_lines = []
        self.edge_labels = []
        self.title = None


    def get_artists(self):
        result = [self.node_markers, self.title]
        result.append(self.edge_lines)
        result.extend(self.node_labels.values())
        result.extend(self.edge_labels)
        return result


    def init_plot(self, linewidth=3):

        G = self.network.graph
        pos = G.nodes.data('position')

        self.axis.clear()
        self.axis.set_axis_off()
        self.node_markers = nx.draw_networkx_nodes(
            G, pos, ax=self.axis, node_size=100,
            node_shape='s', node_color='#8888FF',
        )
        self.node_labels = nx.draw_networkx_labels(
            G, pos, ax=self.axis,
            labels=dict(G.nodes.data('name')),
            font_size=9,
            bbox=dict(facecolor='white', alpha=0.5, edgecolor='none')
        )

        self.edge_lines = nx.draw_networkx_edges(
            G, pos, ax=self.axis, width=linewidth, edge_color='#4444FF',
        )

        labels = {(l, r): v for l, r, v in G.edges.data('label')}
        if self.with_edge_labels:
            self.edge_labels = nx.draw_networkx_edge_labels(
                G, pos, ax=self.axis,
                edge_labels=labels, font_size=7
            )
        else:
            self.edge_labels = []

        self.title = self.axis.text(
            x=1.0, y=1.0, s='...',
            horizontalalignment='right',
            verticalalignment='top',
            transform=self.axis.transAxes
        )
        self.axis.set_xlim(*[x * 1.1 for x in self.axis.get_xlim()])
        plt.axis('off')
        return self.get_artists()

    def update(self, t):
        volume = self.network.get_volumes()  # volume of the nodes in mm/day
        flux = self.network.get_fluxes(t)  # fluxes over edges in mm/day
        self.node_markers.set_sizes(10 * np.sqrt(volume))
        self.edge_lines.set_linewidths(flux)
        if t:
            self.title.set_text(str(t))
        return self.get_artists()

    def get_animator(self, integration):
        """
        Returns a matplotlib.animation.FuncAnimator object that uses
        the integration iteratable to advance your model to animate the HillPlot

        Usage example:

        >>>p=cmf.project()
        >>>solver = cmf.CVodeIntegrator(p, 1e-9)
        >>>fg = Fluxogram(outlet)
        >>>animator = fg.get_animator(solver.run(datetime(2012, 1, 1), datetime(2012, 2, 1), timedelta(hours=1)))

        :param integration: An iterable that advances the model and yields the current time
        :return: A matplotlib.animation.FuncAnimator
        """

        fa = FuncAnimation(self.axis.figure, func=self.update,
                           init_func=self.init_plot,
                           frames=integration, blit=False,
                           interval=1, repeat=False)
        return fa



#!/usr/bin/env python3

import sys


def get_commands():
    """
    Gets a dictionary of allowed commands
    """

    from . import commands
    return {
        n: f
        for n, f in vars(commands).items()
        if not n.startswith('_')
    }


def main(args=sys.argv):

    commands = get_commands()
    if len(args) > 1 and args[1] in commands:
        f = commands[args[1]]
        if len(args) == 2 or args[2] == 'help':
            sys.stderr.write(f.__doc__)
        else:
            f(*args[2:])
    elif len(args) > 1 and args[1] == 'help':
        print('Runs a custom build lumped cmf model')
        for n, f in commands.items():
            print()
            print(f'cmf.lumped {n} ...')
            print(f.__doc__)
    else:
        cmd_list = '|'.join(commands)
        sys.stderr.write(f'Usage: cmf.lumped [{cmd_list}] path/to/model.py')


if __name__ == '__main__':
    main()




#!/usr/bin/env python3
from spotpy import describe
from pathlib import Path
from textwrap import dedent
import cmf
import shutil
import re
import sys
from sphinx.cmd.build import make_main as sphinx_make
from logging import getLogger
logger = getLogger(__name__)


def name(setup):
    return setup.name


def Name(setup):
    return setup.name.capitalize()


def get_doc_class(setup, classname):
    return getattr(sys.modules[setup.__module__], classname, None)


class Implementation:

    def __init__(self, setup):
        self.setup = setup

    def parameter_text(self):
        cls = get_doc_class(self.setup, 'Parameters')
        return (
                dedent(cls.__doc__) + '\n' +
                '\n'.join(
                    f':{p.name}: [{p.minbound:0.4g}..{p.maxbound:0.4g}] {p.description}'
                    for p in self.setup.parameters
                )
        )

    def model_text(self):
        cls = type(self.setup)
        return cls.describe(self.setup)

    def project_text(self):
        return cmf.describe(self.setup.project)

    def __repr__(self):
        return f'Implementation({self.setup})'

    def __str__(self):
        return '\n'.join((self.parameter_text(), self.model_text(), self.project_text()))


class Documentation:

    def copy_conf_py(self):
        """
        Copies the conf.py file needed to compile the documentation from the sources,
        if it is not present in the target directory
        """
        if not (self.homedir / 'conf.py').exists():
            srcdir = Path(__file__).parent
            conf_py = srcdir / 'conf.py'
            shutil.copy(conf_py, self.homedir)

    def __init__(self, setup, homedir: Path = None):
        self.setup = setup
        self.homedir = homedir or Path(f'{name(self.setup)}-docs')
        self.homedir.mkdir(parents=True, exist_ok=True)

    def copy_figures(self, text):
        figures = re.findall(r'\.\. figure:: (.*)', text)
        for fig in figures:
            if (img := self.homedir.parent / fig).exists():
                shutil.copy(img, self.homedir)

    def write_doc_text(self, classname: str):

        if cls := get_doc_class(self.setup, classname):
            doc = cls.describe(self.setup)
            self.write_text(classname.lower(), doc)
            self.copy_figures(doc)
        else:
            logger.warning(f'No class <{classname}> found in {self.setup.name}')

    def write_text(self, filename: str, text: str, module_prefix=True):
        if module_prefix:
            path = self.homedir / f'{self.setup.name}.{filename}.rst'
        else:
            path = self.homedir / f'{filename}.rst'

        with path.open('w', encoding='utf-8') as f:
            f.write(str(text))

    def index_text(self):
        """
        Uses either an existing index.rst, or copies ../index.rst to the current source directory and extends it
        with a link to the current model. If neither exists, a base index.rst is created.
        """
        if (self.homedir / 'index.rst').exists():
            return (self.homedir / 'index.rst').read_text(encoding='utf-8')
        elif (self.homedir.parent / 'index.rst').exists():
            return (self.homedir.parent / 'index.rst').read_text(encoding='utf-8') + \
                   f'\n   {Name(self.setup)} <{name(self.setup)}.main>'
        else:
            return dedent(f'''
                Modelling results
                ==============================

                .. toctree::
                   :maxdepth: 2

                   {Name(self.setup)} <{name(self.setup)}.main>

                ''')

    def main_text(self):
        toctree = '.. toctree::\n   ' + '\n   '.join(
            f'{name(self.setup)}.{chapter}'
            for chapter in ('concept', 'implementation', 'result', 'discussion', 'bibliography')
        )
        mod = sys.modules.get(self.setup.name, None)
        mod_doc = getattr(mod, '__doc__', '') or ''
        return mod_doc + '\n' + toctree

    def make_rst(self):
        self.write_doc_text('Concept')
        self.write_text('implementation', str(Implementation(self.setup)))
        self.write_doc_text('Result')
        self.write_doc_text('Discussion')
        self.write_doc_text('Bibliography')
        self.write_text('main', self.main_text())
        self.write_text('index', self.index_text(), module_prefix=False)
        self.copy_conf_py()
        return self

    def compile_html(self):
        """
        Erstellt aus den rst-Dateien die html-Dokumentation
        """
        builddir = self.homedir / '_build'
        args = ['-M', 'html', str(self.homedir), str(builddir)]
        sphinx_make(args)
        return builddir






from spotpy.gui.mpl import Widget, as_scalar, ValueChanger
import numpy as np
from matplotlib.widgets import Slider, Button
from matplotlib import pylab as plt
from matplotlib.animation import FuncAnimation
import time
from spotpy.parameter import get_parameters_array, create_set
from ..basemodel import BaseModel
import cmf
import logging

from .fluxogram import Fluxogram

logger = logging.getLogger(__name__)
class GUI:

    def __init__(self, setup):
        """
        Creates the GUI

        :param setup: A spotpy setup
        """

        logger.debug('init gui')
        self.fig = plt.figure(type(setup).__name__)
        self.time_ax = plt.axes([0.05, 0.1, 0.9, 0.4])

        self.flux_ax = plt.axes([0.05, 0.55, 0.5, 0.4], facecolor='0.8')
        self.fluxogram = Fluxogram(setup.outlet, self.flux_ax)
        self.fluxogram.init_plot(3)
        self.button_play = Widget([0.85, 0.01, 0.04, 0.03], Button, '\u25B6', on_clicked=self.animate)
        # self.button_stop = Widget([0.8, 0.01, 0.04, 0.03], Button, '\u25A0', on_clicked=self.stop)
        self.button_simulate = Widget([0.9, 0.01, 0.04, 0.03], Button, '\u25B6\u25B6', on_clicked=self.run)
        # self.button_clear = Widget([0.9, 0.01, 0.04, 0.03], Button, '\u2718', on_clicked=self.clear)
        self.setup: BaseModel = setup
        self.parameter_values = {
            p['name']: p['optguess']
            for p in get_parameters_array(self.setup)
        }
        self.sliders = self._make_widgets()
        self.lines = []
        self.running = False
        self.clear()


    def close(self):
        plt.close(self.fig)

    def __enter__(self):
        logger.debug('enter gui context')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @staticmethod
    def show():
        """
        Calls matplotlib.pylab.show to show the GUI.
        """
        plt.show()

    def _make_widgets(self):
        """
        Creates the sliders
        :return:
        """
        if hasattr(self, 'sliders'):
            for s in self.sliders:
                s.ax.remove()

        sliders = []
        params = get_parameters_array(self.setup)

        def calc_step():
            h = self.flux_ax.get_position().height
            n = len(params)
            step = h / n
            return max(0.005, min(0.05, step))

        step = calc_step()
        for i, value in enumerate(params):
            rect = [0.75, 1.0 - step * (i + 1), 0.2, step - 0.005]
            s = Widget(rect, Slider, params['name'][i], params['minbound'][i], params['maxbound'][i],
                       valinit=params['optguess'][i], on_changed=ValueChanger(params['name'][i], self.parameter_values))
            sliders.append(s)
        plt.draw()
        return sliders

    def get_parameter_set(self):
        return create_set(self.setup, **self.parameter_values)

    def clear(self, _=None):
        """
        Clears the graph and plots the evalution
        """
        obs = self.setup.evaluation()
        day = np.timedelta64(1, 'D')
        begin = np.datetime64(self.setup.begin)
        time_dim = begin + day * np.arange(len(obs))

        self.time_ax.clear()

        self.lines = list(
            self.time_ax.plot(time_dim, obs, 'k:', label='Observation', zorder=2)
        )
        self.time_ax.legend()

    def run(self, _=None):
        """
        Runs the model and plots the result
        """
        logger.info('silent run')
        self.running = False
        self.time_ax.set_title('Calculating...')
        plt.draw()
        time.sleep(0.001)

        parset = create_set(self.setup, **self.parameter_values)
        logger.debug('start simulation')
        sim = np.array(self.setup.simulation(parset))
        logger.debug('simulation done')
        objf = as_scalar(self.setup.objectivefunction(sim, self.setup.evaluation()))
        day = np.timedelta64(1, 'D')
        begin = np.datetime64(self.setup.begin)
        time_dim = begin + day * np.arange(len(sim))
        self.lines.extend(self.time_ax.plot(time_dim, sim, '-'))
        self.time_ax.legend()
        self.time_ax.set_title(type(self.setup).__name__)
        self.time_ax.figure.canvas.draw_idle()
        logger.debug('plot result')

    def stop(self, _=None):
        logger.info('pressed stop')
        self.running = False

    def animate(self, _=None):
        logger.info('animation')
        self.running = False
        def cmf_time_to_ms(t):
            return (t - cmf.Time(1, 1, 1970)).AsMilliseconds()
        outflow = [self.setup.output(self.setup.begin)]
        timeline_ms = [cmf_time_to_ms(cmf.AsCMFtime(self.setup.begin))]

        def update(t):
            outflow.append(self.setup.output(t))
            timeline_ms.append(cmf_time_to_ms(t))
            flux_artists = self.fluxogram.update(t)
            time_dim = np.array(timeline_ms, dtype='datetime64[ms]')
            self.lines[-1].set_data(time_dim, outflow)
            return self.lines + flux_artists

        def init():
            act_line, = self.time_ax.plot([], [], '-', label='actual run')
            self.lines.append(act_line)
            self.time_ax.legend()

            flux_artists = self.fluxogram.init_plot(0)
            return self.lines + flux_artists

        def stoppable(it):
            t = self.setup.begin
            try:
                while self.running:
                    t = next(iter(it))
                    yield t
            except StopIteration:
                pass
            finally:
                time_dim = np.array(timeline_ms, dtype='datetime64[ms]')
                self.lines[-1].set_data(time_dim, outflow)
                self.fluxogram.init_plot(0)
                self.fluxogram.update(t)
                for l in self.lines:
                    self.time_ax.figure.draw(l)
                self.time_ax.figure.canvas.draw_idle()
                self.flux_ax.figure.canvas.draw_idle()


        self.running = True
        parset = self.get_parameter_set()
        iterator = stoppable(self.setup.iterate(parset))
        logger.debug('create animation')
        self.animator = FuncAnimation(
            self.fig, update,
            iterator,
            init_func=init, repeat=False,
            interval=1, blit=True
        )







# -*- coding: utf-8 -*-
"""
Ein 2-Speicher lumped Modell für das Einzugsgebiet der Fulda
===============================================================

:author: Philipp Kraft, Mat-Nr: xxxxx-xxxxx



"""
import cmf
import spotpy
from spotpy.parameter import Uniform, Constant
import numpy as np
import datetime
import importlib.util
import os
import sys
from .dataprovider import DataProvider
from .doctools import DocClass

def load_module_from_path(path_to_module:str):
    """
    Loads a module from a path
    :param path_to_module:
    :return:

    See: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    """
    name = os.path.basename(path_to_module).replace('.py', '')
    spec = importlib.util.spec_from_file_location(name, path_to_module)
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    spec.loader.exec_module(module)

    return module


def get_model_class(path_to_module:str, classname=None):
    """
    This function get the class of a lumped cmf model from
    a Python file
    :param path_to_module: The file path to the module
    :param classname: A classname, if None the first childclass of BaseModel is returned
    :return: The model class
    """

    module = load_module_from_path(path_to_module)
    if classname:
        return getattr(module, classname)
    else:
        # Loop through all members of the module and detect a class that is derived
        # from the cmflumped.BaseModel
        for name, obj in vars(module).items():
            if (
                    name[0] != '_'
                    and isinstance(obj, type)
                    and obj is not module.BaseModel
                    and issubclass(obj, module.BaseModel)
            ):
                return obj
        # Raise Error if no fitting class is found
        raise ValueError(f'Module "{path_to_module}" has no class that derives from cmflumped.BaseModel')


def u(vmin, vmax, default=None, doc=None):
    """
    Creates a uniform distributed parameter
    :param vmin: Minimum value
    :param vmax: Maximum value
    :param default: Default value
    :param doc: Documentation of the parameter (use reStructuredText)
    :return: The parameter object
    """
    if default is None:
        default = 0.5 * (vmin + vmax)
    return Uniform(vmin, vmax, optguess=default, doc=doc, minbound=vmin, maxbound=vmax)


def constant(vmin, vmax, default=None, doc=None):
    if default is None:
        default = 0.5 * (vmin + vmax)
    return Constant(default, optguess=default, minbound=vmin, maxbound=vmax, doc=doc)


class BaseParameters:

    def __get__(self, instance, owner):
        # A magic method for simple use of this object
        return spotpy.parameter.get_parameters_from_setup(self)

    @classmethod
    def to_string(cls):
        params = cls()
        return cls.__doc__ + '\n'.join(
            f':{p.name}: [{p.minbound:0.4g}..{p.maxbound:0.4g}] {p.description}'
            for p in spotpy.parameter.get_parameters_from_setup(params)
        )

class BaseModel(DocClass):
    """
    The template for
    """

    verbose = True
    calibration_start = 2000
    validation_start = 2010

    def __init__(self, dataprovider: DataProvider, name: str = None):
        """
        Creates the basic structure of the model
        """
        self.data = dataprovider
        self.name = name or self.__module__
        self.project = cmf.project()
        self.cell: cmf.Cell = self.project.NewCell(0, 0, 0, 1000)

        self.create_nodes()
        self.data.add_stations(self.project)
        self.create_connections(spotpy.parameter.create_set(self, 'optguess'))

# OVERRIDE ->
    def create_nodes(self):
        """
        Create the nodes (storages, distribution nodes and boundaries) of your model.

        eg.

        .. code-block::

            self.soil, self.gw = self.add_layers(1.0, 1.0)
            self.outlet = self.project.NewOutlet('outlet' , 0, 0, 0)

        or something else
        """
        ...

    def create_connections(self, p):
        ...

    def initial_values(self, p):
        """
        Is called before a simulation starts and should reset
        every storage of the model and needs to be defined for each model
        :param p: The parameters object. May or may not be used in the function
        :return: None
        """
        ...

    def create_result_structure(self):
        outflow = cmf.timeseries(self.data.begin, cmf.day)
        outflow.add(self.outlet(self.data.begin))
        return outflow

    def fill_result_structure(self, result, t):
        q = self.output(t)
        result.add(q)
        if self.verbose:
            print(f'{t!s:>12s} Q={q:10.5g}mm/day')
# <- OVERRIDE
    def add_layers(self, *thickness):
        return tuple(
            self.cell.add_layer(d)
            for d in np.cumsum(thickness)
        )

    def iterate(self, p = None):
        """
        Calls `create_connections` and `inital_values` to shape
        the model using the parameter set `p` and returns an
        iterator that advances the model over the whole data period
        """
        self.create_connections(p)
        self.initial_values(p)

        solver = cmf.CVodeIntegrator(self.project, 1e-9)
        solver.use_OpenMP = False
        return solver.run(self.data.begin, self.data.end, cmf.day)

    def output(self, t):
        """
        Defines what the ouput of the model is
        :param t: Time step of the model
        :return: A value representing the model output
        """
        return self.outlet.waterbalance(t)

    def __str__(self):
        return self.name

    def simulation(self, vector):
        """
        This function is only important for spotpy,
        otherwise it is equivalent with "run"
        :param vector:
        :return:
        """
        result = self.create_result_structure()
        duration = datetime.datetime.now()

        for t in self.iterate(vector):
            self.fill_result_structure(result, t)
        result_array = np.array(result)
        if self.verbose:
            print('objective: NSE_c={:0.4g}, NSE_v={:0.4g}, PBIAS_c={:0.4g}, PBIAS_v={:0.4g}'.format(*self.objectivefunction(result_array, self.evaluation())))
            print('duration:', datetime.datetime.now() - duration)
        return result_array

    def evaluation(self):
        """
        Returns the evaluation data
        """
        runoff = self.data.Q
        return np.array(runoff)

    def objectivefunction(self, simulation, evaluation):
        """
        Calculates the goodness of the simulation

        Calculates
         - :math:`NSE_c`: the Nash-Sutcliffe Efficiancy
            for the calibration period (self.begin:self.end)
         - :math:`PBIAS_c`: the procentual bias between model and observation
            for the calibration period (self.begin:self.end)
         - :math:`NSE_v`: the Nash-Sutcliffe Efficiancy
            for the validation period (self.end:self.data.end)
         - :math:`PBIAS_v`: the procentual bias between model and observation
            for the validation period (self.end:self.data.end)

        and returns these objectives as a list in that order
        """

        def getindex(year):
            """
            Returns the array position for Jan 1st of the respective year
            :param year: a year
            :return: int
            """
            dt = datetime.datetime(year, 1, 1)
            return (dt - self.data.begin).days

        c_start = getindex(self.calibration_start)
        v_start = getindex(self.validation_start)
        nse_c = spotpy.objectivefunctions.nashsutcliffe(evaluation[c_start:v_start], simulation[c_start:v_start])
        pbias_c = spotpy.objectivefunctions.pbias(evaluation[c_start:v_start], simulation[c_start:v_start])
        nse_v = spotpy.objectivefunctions.nashsutcliffe(evaluation[v_start:], simulation[v_start:])
        pbias_v = spotpy.objectivefunctions.pbias(evaluation[v_start:], simulation[v_start:])

        return [nse_c, nse_v, pbias_c, pbias_v]


    @property
    def begin(self):
        return self.data.begin

    @property
    def end(self):
        return self.data.end






import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as fh:
    requirements = fh.readlines()


def get_version(rel_path):
    import os
    def read(rel_path):
        here = os.path.abspath(os.path.dirname(__file__))
        with open(os.path.join(here, rel_path)) as f:
            return f.read()
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name="cmflumped",
    version=get_version('cmflumped/__init__.py'),
    author="Philipp Kraft",
    author_email="philipp.kraft@envr.jlug.de",
    description="A toolbox to build, debug, run and calibrate lumped cmf models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/philippkraft/cmf.lumped",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=requirements,
    entry_points = {
        'console_scripts': [
            'cmf.lumped=cmflumped.__main__:main'
        ]
    }
)


import spotpy
import spotpy.describe
import os


def parallel():
    """
    Returns 'mpi', if this code runs with MPI, else returns 'seq'
    :return:
    """
    return 'mpi' if 'OMPI_COMM_WORLD_SIZE' in os.environ else 'seq'


def get_runs(default=1):
    """
    Returns the number of runs, given by commandline or variable
    :param default: Return this if no other source for number of runs has been found
    :return: int
    """
    # Get number of runs
    if 'SPOTPYRUNS' in os.environ:
        # from environment
        return int(os.environ['SPOTPYRUNS'])
    else:
        # run once
        return default


def sample(model, runs, algname='lhs', save_threshold=None):
    runs = get_runs(runs)
    alg = getattr(spotpy.algorithms, algname)
    sampler = alg(
        model, sim_timeout=600,
        dbname=str(model), dbformat='hdf5', parallel=parallel(), save_threshold=save_threshold)
    print(spotpy.describe.sampler(sampler))
    print(spotpy.describe.setup(model))
    sampler.sample(runs)



from textwrap import dedent


class DocFunction:
    def __init__(self, template, **kwargs):
        self.kwargs: dict = kwargs
        self.template = dedent(template)

    def __format__(self, format_spec):
        kwargs = self.kwargs.copy()
        args = []
        pair: str
        for item in format_spec.split(';'):
            if '=' in item:
                k, v = item.split('=', 1)
                kwargs[k.strip()] = v.strip()
            elif item:
                args.append(item.strip())

        for k, a in zip(self.kwargs, args):
            kwargs[k] = a

        return self.template.format(**kwargs)


class DocClass:
    figure = DocFunction(
        """
        .. _fig-{module}-{name}:

        .. figure:: {module}.{name}.{ext}
            :width: 800px
            :align: center
            :figclass: align-center

            {description}
        """, name='', description='', module='', ext='png')

    fig_ref = DocFunction(':numref:`fig-{module}-{name}`', name='')

    bib_ref = DocFunction('[{module}-{shortcut}]_', shortcut='')

    bib_item = DocFunction('.. [{module}-{shortcut}] {title} {doi}', shortcut='', doi='', title='')

    Class = DocFunction("""
    .. autoclass:: {module}.{classname}
        :members:
    """)

    @classmethod
    def describe(cls, setup=None, **kwargs) -> str:
        template = dedent(cls.__doc__)
        module = setup and setup.name
        classname = setup and setup.__class__.__name__
        docfunctions = {
            k: getattr(cls, k) for k in dir(cls)
            if isinstance(getattr(cls, k), DocFunction)
        }

        for dcf in docfunctions.values():
            dcf.kwargs['module'] = module
            dcf.kwargs['classname'] = classname
            dcf.kwargs['setup'] = setup

        return template.format(setup=setup, module=module, classname=classname, **docfunctions, **kwargs)



import cmf
import pandas as pd
import datetime as dt

class DataProvider:
    """
    Holds the forcing and calibration data
    """

    def __init__(self, data, Q='Q', P='P', E='ETpot', Tmin=None, Tmax=None):
        """
        Makes a new dataprovider for a lumped cmf model

        :param data: A pandas dataframe with the dates as the index
        :param Q: Name or index of the dataframe's column containing the validation discharge in mm/day
        :param P: Name or index of the dataframe's column containing Precipitation in mm/day
        :param E: Name or index of the dataframe's column containing potential Evaporation (ETpot) in mm/day
        :param Tmin: Name or index of the dataframe's column containing the daily min Temperature (°C)
                    - None means no Temperature is given
        :param Tmax: Name or index of the dataframe's column containing the daily min Temperature (°C)
                    - None means: use Tmin as daily average Temperature
        """
        # Get begin, step and end from the date column
        self.begin: dt.datetime = data.index[0].to_pydatetime()
        self.end: dt.datetime = data.index[-1].to_pydatetime()
        self.step: dt.timedelta = data.index[1].to_pydatetime() - self.begin

        def a2ts(a):
            """Converts an array column to a timeseries"""
            return cmf.timeseries.from_array(self.begin, self.step, a)

        def get_col(c):
            if type(c) is int:
                return data[data.columns[c]]
            else:
                return data[c]

        self.P = a2ts(get_col(P))
        self.Q = a2ts(get_col(Q))
        self.ETpot = a2ts(get_col(E))

        self.Tmin = a2ts(get_col(Tmin)) if Tmin else None
        self.Tmax = a2ts(get_col(Tmax)) if Tmax else None


    def add_stations(self, project):
        """
        Creates a rainstation and a meteo station for the cmf project
        :param project: A cmf.project
        :return: rainstation, meteo
        """
        rainstation = project.rainfall_stations.add('Glauburg', self.P, (0, 0, 0))

        project.use_nearest_rainfall()

        # Temperaturdaten
        meteo = project.meteo_stations.add_station('Glauburg', (0, 0, 0))
        if self.Tmin:
            meteo.Tmin = self.Tmin
        else:
            meteo.Tmin = cmf.timeseries.from_scalar(10.0)
        if self.Tmax:
            meteo.Tmax = self.Tmax
        else:
            meteo.Tmax = meteo.Tmin

        project.use_nearest_meteo()

        return rainstation, meteo

    def summerize(self):
        for ts_name in 'P ETpot T Q'.split():
            ts = getattr(self, ts_name)
            print(ts_name, cmf.describe(ts).replace('\n', ''))


def load_csv(csv_file: str, date=0, Q='Q', P='P', E='ETpot', Tmin=None, Tmax=None, **kwargs) -> DataProvider:
    """
    Loads driver and calibration
    -----------------------------

    :param csv_file: The csv file with decimal point and commas as list seperator
    :param date: Name or index of the dataframe's column containing dates
    :param Q: Name or index of the dataframe's column containing the validation discharge in mm/day
    :param P: Name or index of the dataframe's column containing Precipitation in mm/day
    :param E: Name or index of the dataframe's column containing potential Evaporation (ETpot) in mm/day
    :param Tmin: Name or index of the dataframe's column containing the daily min Temperature (°C)
                - None means no Temperature is given
    :param Tmax: Name or index of the dataframe's column containing the daily min Temperature (°C)
                - None means: use Tmin as daily average Temperature

    :param kwargs: Keyword arguments to be passed on to pandas.load_csv

    :return: A DataProvider object

    Example csv file:
    -----------------
    .. code-block::

        date,Prec mm,Temp °C,ET mm,Q mm
        2020-01-01,2.46,5.3,0.2,1.3

    Usage with column names:
    >>> data = load_csv('pteq.csv', date='date', P='Prec mm', E='ET mm', Tmin='Temp °C', Q='Q mm')

    Usage with column positions:
    >>> data = load_csv('pteq.csv', date=0, P=1, E=3, Tmin=2, Q=4)
    """
    data = pd.read_csv(csv_file, index_col=[date], parse_dates=True)
    return DataProvider(data, Q, P, E, Tmin, Tmax)

